/*=========================================================================
 *
 * Reimplementation of the rtkinlinefdk from v1.4. Which mimics a reconstruction
 * during acquisition by splitting the geometry definition and projections adding
 * to one thread and the reconstruction to the other thread. The delay between
 * projections and max number of projections processed per iteration by the
 * reconstruction thread can be set with the ggo parameters
 *
 * Some modification compared to the original were done
 * - Threading in ITK changed with newer versions i'm not familiar with it
 *   so STL thread is now used
 * - Previous implementation used a synchronized inlinefdk which is slower
 *   than just running it async. Now the threads only lock eachother with a
 *   mutex when issues could arrise, otherwise the threads continue independently
 * - default 
 *
 *=========================================================================*/

#include "rtkinlinefdk_ggo.h"
#include "rtkGgoFunctions.h"
#include "rtkConfiguration.h"

#include "rtkThreeDCircularProjectionGeometryXMLFile.h"
#include "rtkDisplacedDetectorForOffsetFieldOfViewImageFilter.h"
#include "rtkParkerShortScanImageFilter.h"
#include "rtkFDKConeBeamReconstructionFilter.h"
#ifdef RTK_USE_CUDA
#  include "rtkCudaDisplacedDetectorImageFilter.h"
#  include "rtkCudaParkerShortScanImageFilter.h"
#  include "rtkCudaFDKConeBeamReconstructionFilter.h"
#endif
#include "rtkFDKWarpBackProjectionImageFilter.h"
#include "rtkCyclicDeformationImageFilter.h"
#include "rtkProgressCommands.h"

#include <itkStreamingImageFilter.h>
#include <itkImageRegionSplitterDirection.h>
#include <itkImageFileWriter.h>

#include <mutex>
#include <thread>

// Pass info between threads
struct ThreadInfoStruct
{
  std::mutex               mutex;
  args_info_rtkinlinefdk * args_info;
  bool                     stop;
  bool                     success;
  unsigned int             nproj;
  double                   radius;
  std::vector<double>      sid;
  std::vector<double>      sdd;
  std::vector<double>      gantryAngle;
  std::vector<double>      projOffsetX;
  std::vector<double>      projOffsetY;
  std::vector<double>      outOfPlaneAngle;
  std::vector<double>      inPlaneAngle;
  std::vector<double>      sourceOffsetX;
  std::vector<double>      sourceOffsetY;
  std::vector<double>      collimationUInf;
  std::vector<double>      collimationUSup;
  std::vector<double>      collimationVInf;
  std::vector<double>      collimationVSup;
  double                   minimumOffsetX;
  double                   maximumOffsetX;
  std::vector<std::string> fileName;
};

void
computeOffsetsFromGeometry(rtk::ThreeDCircularProjectionGeometry::Pointer geometry,
                           double *                                       minOffset,
                           double *                                       maxOffset);

static void
AcquisitionCallback(ThreadInfoStruct * arg);
static void
InlineThreadCallback(ThreadInfoStruct * arg);

int
main(int argc, char * argv[])
{
  GGO(rtkinlinefdk, args_info);

  // initialize the struct we'll pass along
  ThreadInfoStruct * threadInfo = new ThreadInfoStruct;
  threadInfo->nproj = 0;
  threadInfo->stop = false;
  threadInfo->success = false;
  threadInfo->args_info = &args_info;

  std::thread acquisitionThread(AcquisitionCallback, threadInfo);

  // lets wait a bit before starting the reconstruction thread
  std::this_thread::sleep_for(std::chrono::milliseconds(200));
  std::thread reconstructionThread(InlineThreadCallback, threadInfo);

  reconstructionThread.join();

  if (threadInfo->success)
  {
    if (threadInfo->args_info->verbose_flag)
      std::cout << "Successfully finished inline reconstruction" << std::endl;
    delete threadInfo;
    return EXIT_SUCCESS;
  }
  else
  {
    if (threadInfo->args_info->verbose_flag)
      std::cout << "Failed to reconstruct " << std::endl;
    return EXIT_FAILURE;
  }
}


static void
AcquisitionCallback(ThreadInfoStruct * threadInfo)
{
  // setup variables
  double                                         minOffset, maxOffset;
  std::unique_lock<std::mutex>                   lock(threadInfo->mutex);
  std::vector<std::string>                       names = rtk::GetProjectionsFileNamesFromGgo(*(threadInfo->args_info));
  rtk::ThreeDCircularProjectionGeometry::Pointer geometry;
  geometry = rtk::ReadGeometry(threadInfo->args_info->geometry_arg);
  computeOffsetsFromGeometry(geometry, &minOffset, &maxOffset);
  threadInfo->radius = geometry->GetRadiusCylindricalDetector();
  threadInfo->minimumOffsetX = minOffset;
  threadInfo->maximumOffsetX = maxOffset;
  lock.unlock();

  unsigned int nproj = geometry->GetMatrices().size();
  // start the fake acquisition
  for (unsigned int i = 0; i < nproj; i++)
  {
    std::unique_lock<std::mutex> lock(threadInfo->mutex);

    threadInfo->sdd.push_back(geometry->GetSourceToDetectorDistances()[i]);
    threadInfo->sid.push_back(geometry->GetSourceToIsocenterDistances()[i]);
    threadInfo->gantryAngle.push_back(geometry->GetGantryAngles()[i]);
    threadInfo->sourceOffsetX.push_back(geometry->GetSourceOffsetsX()[i]);
    threadInfo->sourceOffsetY.push_back(geometry->GetSourceOffsetsY()[i]);
    threadInfo->projOffsetX.push_back(geometry->GetProjectionOffsetsX()[i]);
    threadInfo->projOffsetY.push_back(geometry->GetProjectionOffsetsY()[i]);
    threadInfo->inPlaneAngle.push_back(geometry->GetInPlaneAngles()[i]);
    threadInfo->outOfPlaneAngle.push_back(geometry->GetOutOfPlaneAngles()[i]);
    threadInfo->collimationUInf.push_back(geometry->GetCollimationUInf()[i]);
    threadInfo->collimationUSup.push_back(geometry->GetCollimationUSup()[i]);
    threadInfo->collimationVInf.push_back(geometry->GetCollimationVInf()[i]);
    threadInfo->collimationVSup.push_back(geometry->GetCollimationVSup()[i]);
    threadInfo->minimumOffsetX = minOffset;
    threadInfo->maximumOffsetX = maxOffset;
    threadInfo->fileName.push_back(names[std::min(i, (unsigned int)names.size() - 1)]);
    threadInfo->nproj = i + 1;
    threadInfo->stop = (i == nproj - 1);

    if (threadInfo->args_info->verbose_flag)
      std::cout << "AcquisitionCallback has simulated the acquisition of projection #" << i << std::endl;

    lock.unlock();
    // we expect the acquisiton to take a bit of time so lets suspend here for a bit note that the
    // acquisiton thread will also occassionally be locked by the mutex so the real delay might be
    // slightly larger but that should be minimal
    std::this_thread::sleep_for(std::chrono::milliseconds(threadInfo->args_info->delay_arg));
  }

  return;
}

static void
InlineThreadCallback(ThreadInfoStruct * threadInfo)
{
  // setup types
  using PixelType = float;
  constexpr unsigned int Dimension = 3;
  using CPUOutputImageType = itk::Image<PixelType, Dimension>;
#ifdef RTK_USE_CUDA
  using OutputImageType = itk::CudaImage<PixelType, Dimension>;
#else
  using OutputImageType = CPUOutputImageType;
#endif
  std::vector<std::string> fileNames;

  // geometry
  rtk::ThreeDCircularProjectionGeometry::Pointer geometry;
  geometry = rtk::ThreeDCircularProjectionGeometry::New();

  // projection reader
  using ReaderType = rtk::ProjectionsReader<OutputImageType>;
  ReaderType::Pointer reader = ReaderType::New();

  // Extract filter to process X projections at a time
  using ExtractFilterType = itk::ExtractImageFilter<OutputImageType, OutputImageType>;
  ExtractFilterType::Pointer extract = ExtractFilterType::New();
  extract->SetInput(reader->GetOutput());
  ExtractFilterType::InputImageRegionType subsetRegion;

  // displaced detector weighting - this one still does not work if i enable it unless i only start reconstruction after
  // many projections have passed (tested it at 40 and this works).
  using DDFCPUType = rtk::DisplacedDetectorImageFilter<OutputImageType>;
  using DDFOFFFOVType = rtk::DisplacedDetectorForOffsetFieldOfViewImageFilter<OutputImageType>;
#ifdef RTK_USE_CUDA
  using DDFType = rtk::CudaDisplacedDetectorImageFilter;
#else
  using DDFType = rtk::DisplacedDetectorForOffsetFieldOfViewImageFilter<OutputImageType>;
#endif
  DDFCPUType::Pointer ddf;
  ddf = DDFOFFFOVType::New();
  ddf->SetInput(extract->GetOutput());
  ddf->SetGeometry(geometry);
  ddf->SetDisable(threadInfo->args_info->nodisplaced_flag);
  ddf->SetOffsets(threadInfo->minimumOffsetX, threadInfo->maximumOffsetX);

  // Create reconstructed image
  using ConstantImageSourceType = rtk::ConstantImageSource<OutputImageType>;
  ConstantImageSourceType::Pointer constantImageSource = ConstantImageSourceType::New();
  rtk::SetConstantImageSourceFromGgo<ConstantImageSourceType, args_info_rtkinlinefdk>(constantImageSource,
                                                                                      *(threadInfo->args_info));

  // This macro sets options for fdk filter because TFFTPrecision is not the same, e.g. for CPU and CUDA
#define SET_FELDKAMP_OPTIONS(f)                                                                                        \
  f->SetInput(0, constantImageSource->GetOutput());                                                                    \
  f->SetInput(1, ddf->GetOutput());                                                                                    \
  f->SetGeometry(geometry);                                                                                            \
  f->GetRampFilter()->SetTruncationCorrection(threadInfo->args_info->pad_arg);                                         \
  f->GetRampFilter()->SetHannCutFrequency(threadInfo->args_info->hann_arg);                                            \
  f->SetProjectionSubsetSize(threadInfo->args_info->batchlimit_arg);                                                   \
  if (threadInfo->args_info->verbose_flag)                                                                             \
  {                                                                                                                    \
    f->AddObserver(itk::AnyEvent(), progressCommand);                                                                  \
  }

  // FDK reconstruction filtering
  using FDKCPUType = rtk::FDKConeBeamReconstructionFilter<OutputImageType>;
  FDKCPUType::Pointer feldkamp;
#ifdef RTK_USE_CUDA
  using FDKCUDAType = rtk::CudaFDKConeBeamReconstructionFilter;
  FDKCUDAType::Pointer feldkampCUDA;
#endif

  itk::Image<PixelType, Dimension> * pfeldkamp = nullptr;
  if (!strcmp(threadInfo->args_info->hardware_arg, "cpu"))
  {
    // Progress reporting
    using PercentageProgressCommandType = rtk::PercentageProgressCommand<FDKCPUType>;
    PercentageProgressCommandType::Pointer progressCommand = PercentageProgressCommandType::New();

    feldkamp = FDKCPUType::New();
    SET_FELDKAMP_OPTIONS(feldkamp);
  }
  else if (!strcmp(threadInfo->args_info->hardware_arg, "cuda"))
  {
#ifdef RTK_USE_CUDA

    using PercentageProgressCommandType = rtk::PercentageProgressCommand<FDKCUDAType>;
    PercentageProgressCommandType::Pointer progressCommand = PercentageProgressCommandType::New();

    feldkampCUDA = FDKCUDAType::New();
    SET_FELDKAMP_OPTIONS(feldkampCUDA);
#else
    std::cerr << "The program has not been compiled with cuda option" << std::endl;
    exit(EXIT_FAILURE);
#endif
  }

  // setup export volume
  using WriterType = itk::ImageFileWriter<CPUOutputImageType>;
  WriterType::Pointer writer = WriterType::New();
  writer->SetFileName(threadInfo->args_info->output_arg);

  geometry->SetRadiusCylindricalDetector(threadInfo->radius);

  if (threadInfo->args_info->verbose_flag)
    std::cout << "Reconstruction thread has entered in the processing loop" << std::endl;

  int          currentSlice = 1;
  unsigned int lastSlice = 0;
  for (;;)
  {
    std::unique_lock<std::mutex> lock(threadInfo->mutex);
    bool                         isLocked = true;
    if (geometry->GetMatrices().size() < threadInfo->nproj)
    {
      if (threadInfo->args_info->verbose_flag)
        std::cout << "Reconstruction thread is processing projection #" << currentSlice << " to #"
                  << threadInfo->nproj - 1 << std::endl;

      for (auto & elem : threadInfo->fileName)
      {
        if (elem != "" && (fileNames.size() == 0 || fileNames.back() != elem))
        {
          fileNames.push_back(elem); // if i understood the reader well enough more filenames than what we select with
                                     // the subregion shouldn't be an issue?
        }
      }

      int batchCount = 1;
      for (lastSlice; lastSlice < threadInfo->nproj - 1; lastSlice++)
      {
        if (batchCount > threadInfo->args_info->batchlimit_arg)
          break;

        ++batchCount;

        geometry->AddProjectionInRadians(threadInfo->sid[lastSlice],
                                         threadInfo->sdd[lastSlice],
                                         threadInfo->gantryAngle[lastSlice],
                                         threadInfo->projOffsetX[lastSlice],
                                         threadInfo->projOffsetY[lastSlice],
                                         threadInfo->outOfPlaneAngle[lastSlice],
                                         threadInfo->inPlaneAngle[lastSlice],
                                         threadInfo->sourceOffsetX[lastSlice],
                                         threadInfo->sourceOffsetY[lastSlice]);
        geometry->SetCollimationOfLastProjection(threadInfo->collimationUInf[lastSlice],
                                                 threadInfo->collimationUSup[lastSlice],
                                                 threadInfo->collimationVInf[lastSlice],
                                                 threadInfo->collimationVSup[lastSlice]);
      }

      lock.unlock(); // here we can already release as we should no longer encounter any race conditions
      isLocked = false;

      if (geometry->GetMatrices().size() < 3)
      {
        std::this_thread::sleep_for(std::chrono::milliseconds(10));
        continue;
      }

      reader->SetFileNames(fileNames);
      TRY_AND_EXIT_ON_ITK_EXCEPTION(reader->UpdateOutputInformation())

      OutputImageType::RegionType region = reader->GetOutput()->GetLargestPossibleRegion();

      if (geometry->GetMatrices().size() - 1 - currentSlice < 1)
      {
        std::this_thread::sleep_for(std::chrono::milliseconds(20));
        continue; // allow acquisition thread to catch up
      }


      if (threadInfo->args_info->verbose_flag)
      {
        // we need to lock every verbose statement not nested inside the lock to avoid mixing characters in the cmd
        // prompt. could move this to a seperate function but for now this is good enough
        std::unique_lock<std::mutex> lock(threadInfo->mutex);
        std::cout << "Reader size : " << region.GetSize()[0] << " " << region.GetSize()[1] << " " << region.GetSize()[2]
                  << std::endl;
        std::cout << "Reader index : " << region.GetIndex()[0] << " " << region.GetIndex()[1] << " "
                  << region.GetIndex()[2] << std::endl;
        lock.unlock();
      }

      // set subset
      subsetRegion = reader->GetOutput()->GetLargestPossibleRegion();
      subsetRegion.SetIndex(Dimension - 1, currentSlice);
      subsetRegion.SetSize(Dimension - 1, geometry->GetMatrices().size() - 1 - currentSlice);
      extract->SetExtractionRegion(subsetRegion);
      extract->UpdateOutputInformation();
      if (threadInfo->args_info->verbose_flag)
      {
        std::unique_lock<std::mutex> lock(threadInfo->mutex);
        std::cout << "Region size : " << subsetRegion.GetSize()[0] << " " << subsetRegion.GetSize()[1] << " "
                  << subsetRegion.GetSize()[2] << std::endl;
        std::cout << "Region index : " << subsetRegion.GetIndex()[0] << " " << subsetRegion.GetIndex()[1] << " "
                  << subsetRegion.GetIndex()[2] << std::endl;

        ExtractFilterType::InputImageRegionType extractRegion = extract->GetOutput()->GetLargestPossibleRegion();

        std::cout << "Extract region size : " << extractRegion.GetSize()[0] << " " << extractRegion.GetSize()[1] << " "
                  << extractRegion.GetSize()[2] << std::endl;
        std::cout << "Extract region index : " << extractRegion.GetIndex()[0] << " " << extractRegion.GetIndex()[1]
                  << " " << extractRegion.GetIndex()[2] << std::endl;
        lock.unlock();
      }

      currentSlice = geometry->GetMatrices().size() - 1;
      if (!strcmp(threadInfo->args_info->hardware_arg, "cpu"))
      {
        TRY_AND_EXIT_ON_ITK_EXCEPTION(feldkamp->Update());
        OutputImageType::Pointer pimg = feldkamp->GetOutput();
        pimg->DisconnectPipeline();
        feldkamp->SetInput(pimg);
        TRY_AND_EXIT_ON_ITK_EXCEPTION(feldkamp->GetOutput()->UpdateOutputInformation())
        TRY_AND_EXIT_ON_ITK_EXCEPTION(feldkamp->GetOutput()->PropagateRequestedRegion());
      }
#ifdef RTK_USE_CUDA
      else if (!strcmp(threadInfo->args_info->hardware_arg, "cuda"))
      {
        TRY_AND_EXIT_ON_ITK_EXCEPTION(feldkampCUDA->Update());
        OutputImageType::Pointer pimg = feldkampCUDA->GetOutput();
        pimg->DisconnectPipeline();
        feldkampCUDA->SetInput(pimg);
        TRY_AND_EXIT_ON_ITK_EXCEPTION(feldkampCUDA->GetOutput()->UpdateOutputInformation());
        TRY_AND_EXIT_ON_ITK_EXCEPTION(feldkampCUDA->GetOutput()->PropagateRequestedRegion());
      }
#endif
      if (threadInfo->args_info->verbose_flag)
      {
        std::unique_lock<std::mutex> lock(threadInfo->mutex);
        std::cout << "Projection #" << subsetRegion.GetIndex(Dimension - 1) << "to "
                  << subsetRegion.GetIndex(Dimension - 1) + subsetRegion.GetSize(Dimension - 1)
                  << " have been processed in reconstruction." << std::endl;
        lock.unlock();
      }
    }


    if (threadInfo->stop &&
        lastSlice == threadInfo->nproj -
                       1) // we only want to enter this when we also processed everything except first and last slice
    {
      // finish up with the remaining projections + first and last frame -> order of previous inlinefdk example

      if (geometry->GetMatrices().size() < threadInfo->nproj)
      {
        geometry->AddProjectionInRadians(threadInfo->sid[threadInfo->nproj - 1],
                                         threadInfo->sdd[threadInfo->nproj - 1],
                                         threadInfo->gantryAngle[threadInfo->nproj - 1],
                                         threadInfo->projOffsetX[threadInfo->nproj - 1],
                                         threadInfo->projOffsetY[threadInfo->nproj - 1],
                                         threadInfo->outOfPlaneAngle[threadInfo->nproj - 1],
                                         threadInfo->inPlaneAngle[threadInfo->nproj - 1],
                                         threadInfo->sourceOffsetX[threadInfo->nproj - 1],
                                         threadInfo->sourceOffsetY[threadInfo->nproj - 1]);
        geometry->SetCollimationOfLastProjection(threadInfo->collimationUInf[threadInfo->nproj - 1],
                                                 threadInfo->collimationUSup[threadInfo->nproj - 1],
                                                 threadInfo->collimationVInf[threadInfo->nproj - 1],
                                                 threadInfo->collimationVSup[threadInfo->nproj - 1]);
      }

      std::vector<unsigned int> indices{ 0, threadInfo->nproj - 1 };

      for (auto & elem : indices)
      {

        OutputImageType::RegionType region = reader->GetOutput()->GetLargestPossibleRegion();

        if (threadInfo->args_info->verbose_flag)
        {
          std::unique_lock<std::mutex> lock(threadInfo->mutex);
          std::cout << "Reader size : " << region.GetSize()[0] << " " << region.GetSize()[1] << " "
                    << region.GetSize()[2] << std::endl;
          std::cout << "Reader index : " << region.GetIndex()[0] << " " << region.GetIndex()[1] << " "
                    << region.GetIndex()[2] << std::endl;
          lock.unlock();
        }

        subsetRegion.SetIndex(Dimension - 1, elem);
        subsetRegion.SetSize(Dimension - 1, 1);

        extract->SetExtractionRegion(subsetRegion);
        // set subset
        extract->UpdateOutputInformation();


        if (threadInfo->args_info->verbose_flag)
        {
          std::unique_lock<std::mutex> lock(threadInfo->mutex);
          std::cout << "Region size : " << subsetRegion.GetSize()[0] << " " << subsetRegion.GetSize()[1] << " "
                    << subsetRegion.GetSize()[2] << std::endl;
          std::cout << "Region index : " << subsetRegion.GetIndex()[0] << " " << subsetRegion.GetIndex()[1] << " "
                    << subsetRegion.GetIndex()[2] << std::endl;

          ExtractFilterType::InputImageRegionType extractRegion = extract->GetOutput()->GetLargestPossibleRegion();

          std::cout << "Extract region size : " << extractRegion.GetSize()[0] << " " << extractRegion.GetSize()[1]
                    << " " << extractRegion.GetSize()[2] << std::endl;
          std::cout << "Extract region index : " << extractRegion.GetIndex()[0] << " " << extractRegion.GetIndex()[1]
                    << " " << extractRegion.GetIndex()[2] << std::endl;
          lock.unlock();
        }
        if (!strcmpi(threadInfo->args_info->hardware_arg, "cpu"))
        {
          TRY_AND_EXIT_ON_ITK_EXCEPTION(feldkamp->Update());
          OutputImageType::Pointer pimg = feldkamp->GetOutput();
          pimg->DisconnectPipeline();
          feldkamp->SetInput(pimg);
          TRY_AND_EXIT_ON_ITK_EXCEPTION(feldkamp->GetOutput()->UpdateOutputInformation());
          TRY_AND_EXIT_ON_ITK_EXCEPTION(feldkamp->GetOutput()->PropagateRequestedRegion());
        }
#ifdef RTK_USE_CUDA
        else if (!strcmpi(threadInfo->args_info->hardware_arg, "cuda"))
        {
          TRY_AND_EXIT_ON_ITK_EXCEPTION(feldkampCUDA->Update());
          OutputImageType::Pointer pimg = feldkampCUDA->GetOutput();
          pimg->DisconnectPipeline();
          feldkampCUDA->SetInput(pimg);
          TRY_AND_EXIT_ON_ITK_EXCEPTION(feldkampCUDA->GetOutput()->UpdateOutputInformation());
          TRY_AND_EXIT_ON_ITK_EXCEPTION(feldkampCUDA->GetOutput()->PropagateRequestedRegion());
        }
#endif
        if (threadInfo->args_info->verbose_flag)
        {
          std::unique_lock<std::mutex> lock(threadInfo->mutex);
          std::cout << "Projection #" << subsetRegion.GetIndex(Dimension - 1)
                    << " has been processed in reconstruction." << std::endl;
          lock.unlock();
        }
      }

      if (!strcmpi(threadInfo->args_info->hardware_arg, "cpu"))
      {
        pfeldkamp = feldkamp->GetOutput();
      }
#ifdef RTK_USE_CUDA
      else if (!strcmpi(threadInfo->args_info->hardware_arg, "cuda"))
      {
        pfeldkamp = feldkampCUDA->GetOutput();
      }
#endif

      // Streaming depending on streaming capability of writer
      using StreamerType = itk::StreamingImageFilter<CPUOutputImageType, CPUOutputImageType>;
      StreamerType::Pointer streamerBP = StreamerType::New();
      streamerBP->SetInput(pfeldkamp);
      streamerBP->SetNumberOfStreamDivisions(
        threadInfo->args_info
          ->divisions_arg); // threadInfo->args_info->divisions_arg // just a random fixed number, no ggo
      itk::ImageRegionSplitterDirection::Pointer splitter = itk::ImageRegionSplitterDirection::New();
      splitter->SetDirection(
        2); // Prevent splitting along z axis. As a result, splitting will be performed along y axis
      streamerBP->SetRegionSplitter(splitter);

      // Write
      if (threadInfo->args_info->verbose_flag)
        std::cout << "Writing volume... " << std::endl;

      TRY_AND_EXIT_ON_ITK_EXCEPTION(itk::WriteImage(streamerBP->GetOutput(), threadInfo->args_info->output_arg));
      threadInfo->success = true;
      if (isLocked)
        lock.unlock();

      return;
    }
  }
}


void
computeOffsetsFromGeometry(rtk::ThreeDCircularProjectionGeometry::Pointer geometry,
                           double *                                       minOffset,
                           double *                                       maxOffset)
{
  double min = itk::NumericTraits<double>::max();
  double max = -1*itk::NumericTraits<double>::max(); // not 100% sure this is correct but the min would return the numeric trait::min() value with negative offsets. Both fail without nodisplaced

  for (unsigned int i = 0; i < geometry->GetProjectionOffsetsX().size(); i++)
  {
   min = std::min(min, geometry->GetProjectionOffsetsX()[i]);
   max = std::max(max, geometry->GetProjectionOffsetsX()[i]);
  }

  *minOffset = min;
  *maxOffset = max;
}
